#!/usr/bin/env python
#coding=utf8

import inspect
import logging
from sjrpc.utils import RpcHandler, pure
from conf import CCConf
from exceptions import AlreadyRegistered, AuthenticationError, RightError

def listed(func):
    func.__listed__ = True
    return func

class CCHandler(RpcHandler):
    '''
    Base class for handlers of CloudControl server.
    '''
    
    def __init__(self, server):
        self._server = server

    def __getitem__(self, name):
        if name.startswith('_'):
            # Filter the private members access:
            raise KeyError('Remote name %s is private.' % repr(name))
        else:
            logging.debug('Called %s.%s' % (self.__class__.__name__, name))
            return super(CCHandler, self).__getitem__(name)

    @pure
    def list_commands(self):
        cmd_list = []

        for attr in dir(self):
            attr = getattr(self, attr, None)
            if getattr(attr, '__listed__', False):
                cmd = {}
                cmd['name'] = attr.__name__
                doc = inspect.getdoc(attr)
                if doc:
                    cmd['description'] = inspect.cleandoc(doc)
                cmd_list.append(cmd)

        return cmd_list


class OnlineCCHandler(CCHandler):

    def on_disconnect(self, conn):
        self._server.unregister(conn)


class HypervisorHandler(OnlineCCHandler):
    '''
    Handler binded to 'node' role.
    '''
    pass


class ClientHandler(OnlineCCHandler):
    '''
    Handler binded to 'cli' role.
    '''

    @listed
    def list(self, conn, query):
        '''
        List all objects registered on this instance.
        '''

        logging.debug('Executed list function with query %s' % query)
        return self._server.list(query)

    def _vm_action(self, query, method, *args, **kwargs):
        vms = self._server.list(query + '$vm$role')
        hypervisors = list(self._server.iter_connected_role('hypervisor'))
        for hv in hypervisors:
            vm_to_start = []
            for vm in vms:
                if vm['role'] != 'vm':
                    pass
                elif vm['id'].split('.')[0] == hv.login:
                    vm_to_start.append(vm['vm'])
            if vm_to_start:
                hv.connection.call(method, vm_to_start, *args, **kwargs)

    @listed
    def start(self, conn, query):
        self._vm_action(query, 'start_vm')

    @listed
    def stop(self, conn, query, force=False):
        self._vm_action(query, 'stop_vm', force)

    @listed
    def destroy(self, conn, query):
        self.stop(query, force=True)

    @listed
    def pause(self, conn, query):
        self._vm_action(query, 'suspend_vm')

    @listed
    def resume(self, conn, query):
        self._vm_action(query, 'resume_vm')

    @listed
    def passwd(self, conn, query, password, method='ssha'):
        '''
        Define a new password for specified user.
        '''

        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.set_password(obj['a'], password, method)

    @listed
    def addaccount(self, conn, login, role, password=None):
        '''
        Create a new account with specified login.
        '''
        self._server.conf.create_account(login, role, password)

    @listed
    def addtag(self, conn, query, tag_name, tag_value):
        '''
        Add a tag to the account which match the specified query.
        '''
        
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.add_tag(obj['a'], tag_name, tag_value)

    @listed
    def deltag(self, conn, query, tag):
        '''
        Remove a tag of the account with specified login.
        '''
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.remove_tag(obj['a'], tag)

    @listed
    def tags(self, conn, query):
        '''
        Return all static tags attached to the specified login.
        '''
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        tags = {}
        for obj in objects:
            tags[obj['a']] = self._server.conf.show(obj['a'])['tags']
        return tags

    @listed
    def delaccount(self, conn, query):
        '''
        Delete the account with specified login.
        '''
        
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.remove_account(obj['a'])

    @listed
    def close(self, conn, query):
        '''
        Close an account without deleting it.
        '''
        
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.set_enabled(obj['a'], enable=False)

    @listed
    def declose(self, conn, query):
        '''
        Re-open an closed account.
        '''
        
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.set_enabled(obj['a'], enable=True)

    @listed
    def kill(self, conn, query):
        '''
        Disconnect all connected accounts selected by query.
        '''

        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.kill(obj['a'])

    @listed
    def rights(self, conn, query):
        '''
        Get the rights of an object set.
        '''
        
        objects = self._server.list(query)
        rules = {}
        for obj in objects:
            if 'a' in obj:
                rules[obj['a']] = self._server.conf.show(obj['a'])['rights']
            else:
                raise BadObjectError('All objects must have the "a" tag.')
        
        return rules

    @listed
    def addright(self, conn, query, tql, method=None, allow=True, index=None):
        '''
        Add a right rule to the selected objects.
        '''
        
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.add_right(obj['a'], tql, method, allow, index)

    @listed
    def delright(self, conn, query, index):
        '''
        Remove a right rule from the selected objects.
        '''
        
        objects = self._server.list(query)
        for obj in objects:
            if 'a' not in obj:
                raise BadObjectError('All objects must have the "a" tag.')
        for obj in objects:
            self._server.conf.remove_right(obj['a'], index)

    def proxy_client(self, conn, login, command, *args, **kwargs):
        client = self._server.get_connection(login)
        return client.connection.call(command, *args, **kwargs)


class WelcomeHandler(CCHandler):
    '''
    Default handler used on client connections of the server.

    :cvar ROLES: role name/handler mapping
    '''
    
    ROLES = {
        'client': ClientHandler,
        'hypervisor': HypervisorHandler,
    }

    @listed
    def authentify(self, conn, login, password):
        '''
        Authenticate the client.
        '''
        try:
            role = self._server.conf.authentify(login, password)
        except CCConf.UnknownAccount:
            raise AuthenticationError('Unknown login')

        if role is None:
            raise AuthenticationError('Bad login/password')
        else:
            # If authentication is a success, try to register the client:
            try:
                self._server.register(login, role, conn)
            except AlreadyRegistered:
                raise AuthenticationError('Already connected')
            conn.set_handler(WelcomeHandler.ROLES.get(role)(self._server))
            
            return role
